(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typed_value.ml,v 1.10 2014/01/10 19:22:10 fclement Exp $ *)

open Value_type_def;;
open Value_type;;

type native_int = Nativeint.t
and int32 = Int32.t
and int64 = Int64.t
;;

type t =
   | Scalar of scalar
   | Vector of vector
   | Matrix of matrix

and scalar =
  | Sbool of bool
  | Sbool_couple of (bool * bool)
  | Sbool_triple of (bool * bool * bool)
  | Sstring of string
  | Sint of int
  | Sint_couple of (int * int)
  | Sint_triple of (int * int * int)
  | Snative_int of native_int
  | Snative_int_couple of (native_int * native_int)
  | Snative_int_triple of (native_int * native_int * native_int)
  | Sint32 of int32
  | Sint32_couple of (int32 * int32)
  | Sint32_triple of (int32 * int32 * int32)
  | Sint64 of int64
  | Sint64_couple of (int64 * int64)
  | Sint64_triple of (int64 * int64 * int64)
  | Sfloat of float
  | Sfloat_couple of (float * float)
  | Sfloat_triple of (float * float * float)

and vector =
  | Vbool of bool array
  | Vbool_couple of (bool * bool) array
  | Vbool_triple of (bool * bool * bool) array
  | Vstring of string array
  | Vint of int array
  | Vint_couple of (int * int) array
  | Vint_triple of (int * int * int) array
  | Vnative_int of native_int array
  | Vnative_int_couple of (native_int * native_int) array
  | Vnative_int_triple of (native_int * native_int * native_int) array
  | Vint32 of int32 array
  | Vint32_couple of (int32 * int32) array
  | Vint32_triple of (int32 * int32 * int32) array
  | Vint64 of int64 array
  | Vint64_couple of (int64 * int64) array
  | Vint64_triple of (int64 * int64 * int64) array
  | Vfloat of float array
  | Vfloat_couple of (float * float) array
  | Vfloat_triple of (float * float * float) array

and matrix =
  | Mcfloat of float Umatrix.t
  | Mffloat of float Umatrix.t
(** The typed-values that are send in the communications.
    Those values are explicitely tagged with their types. *)
;;

let rec type_of_value = function
  | Scalar s -> scalar_type (type_of_scalar s)
  | Vector v -> vector_type (type_of_vector v)
  | Matrix m -> matrix_type (type_of_matrix m)

and type_of_scalar = function
  | Sbool _ -> bool_type
  | Sbool_couple _ -> bool_couple_type
  | Sbool_triple _ -> bool_triple_type
  | Sstring _ -> string_type
  | Sint _ -> int_type
  | Sint_couple _ -> int_couple_type
  | Sint_triple _ -> int_triple_type
  | Snative_int _ -> native_int_type
  | Snative_int_couple _ -> native_int_couple_type
  | Snative_int_triple _ -> native_int_triple_type
  | Sint32 _ -> int32_type
  | Sint32_couple _ -> int32_couple_type
  | Sint32_triple _ -> int32_triple_type
  | Sint64 _ -> int64_type
  | Sint64_couple _ -> int64_couple_type
  | Sint64_triple _ -> int64_triple_type
  | Sfloat _ -> float_type
  | Sfloat_couple _ -> float_couple_type
  | Sfloat_triple _ -> float_triple_type

and type_of_vector = function
  | Vbool _ -> bool_type
  | Vbool_couple _ -> bool_couple_type
  | Vbool_triple _ -> bool_triple_type
  | Vstring _ -> string_type
  | Vint _ -> int_type
  | Vint_couple _ -> int_couple_type
  | Vint_triple _ -> int_triple_type
  | Vnative_int _ -> native_int_type
  | Vnative_int_couple _ -> native_int_couple_type
  | Vnative_int_triple _ -> native_int_triple_type
  | Vint32 _ -> int32_type
  | Vint32_couple _ -> int32_couple_type
  | Vint32_triple _ -> int32_triple_type
  | Vint64 _ -> int64_type
  | Vint64_couple _ -> int64_couple_type
  | Vint64_triple _ -> int64_triple_type
  | Vfloat _ -> float_type
  | Vfloat_couple _ -> float_couple_type
  | Vfloat_triple _ -> float_triple_type

and type_of_matrix = function
  | Mcfloat _
  | Mffloat _ -> float_type;;

exception Wrong_communication of string;;

let to_string_msg = function
  | Scalar (Sbool _) -> "bool"
  | Scalar (Sbool_couple _) -> "bool_couple"
  | Scalar (Sbool_triple _) -> "bool_triple"
  | Scalar (Sstring _) -> "string"
  | Scalar (Sint _) -> "int"
  | Scalar (Sint_couple _) -> "int_couple"
  | Scalar (Sint_triple _) -> "int_triple"
  | Scalar (Snative_int _) -> "native_int"
  | Scalar (Snative_int_couple _) -> "native_int_couple"
  | Scalar (Snative_int_triple _) -> "native_int_triple"
  | Scalar (Sint32 _) -> "int32"
  | Scalar (Sint32_couple _) -> "int32_couple"
  | Scalar (Sint32_triple _) -> "int32_triple"
  | Scalar (Sint64 _) -> "int64"
  | Scalar (Sint64_couple _) -> "int64_couple"
  | Scalar (Sint64_triple _) -> "int64_triple"
  | Scalar (Sfloat _) -> "float"
  | Scalar (Sfloat_couple _) -> "float_couple"
  | Scalar (Sfloat_triple _) -> "float_triple"
  | Vector (Vbool _) -> "bool_vector"
  | Vector (Vbool_couple _) -> "bool_couple_vector"
  | Vector (Vbool_triple _) -> "bool_triple_vector"
  | Vector (Vstring _) -> "string_vector"
  | Vector (Vint _) -> "int_vector"
  | Vector (Vint_couple _) -> "int_couple_vector"
  | Vector (Vint_triple _) -> "int_triple_vector"
  | Vector (Vnative_int _) -> "native_int_vector"
  | Vector (Vnative_int_couple _) -> "native_int_couple_vector"
  | Vector (Vnative_int_triple _) -> "native_int_triple_vector"
  | Vector (Vint32 _) -> "int32_vector"
  | Vector (Vint32_couple _) -> "int32_couple_vector"
  | Vector (Vint32_triple _) -> "int32_triple_vector"
  | Vector (Vint64 _) -> "int64_vector"
  | Vector (Vint64_couple _) -> "int64_couple_vector"
  | Vector (Vint64_triple _) -> "int64_triple_vector"
  | Vector (Vfloat _) -> "float_vector"
  | Vector (Vfloat_couple _) -> "float_couple_vector"
  | Vector (Vfloat_triple _) -> "float_triple_vector"
  | Matrix (Mcfloat _) -> "c_float_matrix"
  | Matrix (Mffloat _) -> "fortran_float_matrix"
;;

let raise_wrong_communication s t =
  raise
    (Wrong_communication
       (Printf.sprintf "waiting for a %s, received a %s"
          s (to_string_msg t)))
;;

let to_bool = function
  | Scalar s ->
    begin match s with
    | Sbool b -> b
                    | Sbool_couple _ | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "bool" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "bool" t

let to_bool_couple = function
  | Scalar s ->
    begin match s with
    | Sbool_couple bc -> bc
    | Sbool _                              | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "bool_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "bool_couple" t

let to_bool_triple = function
  | Scalar s ->
    begin match s with
    | Sbool_triple bt -> bt
    | Sbool _       | Sbool_couple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "bool_triple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "bool_triple" t

and to_string = function
  | Scalar s ->
    begin match s with
    | Sstring s -> s
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "string" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "string" t

and to_int = function
  | Scalar s ->
    begin match s with
    | Sint i -> i
    | Sstring _
    | Sbool _       | Sbool_couple _       | Sbool_triple _
                    | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int" t

and to_int_couple = function
  | Scalar s ->
    begin match s with
    | Sint_couple ic -> ic
    | Sstring _
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sint _                               | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int_couple" t

and to_int_triple = function
  | Scalar s ->
    begin match s with
    | Sint_triple it -> it
    | Sstring _
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sint _        | Sint_couple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int_triple" (Scalar s) end
  | Vector _ | Matrix _ as t -> raise_wrong_communication "int_triple" t

and to_native_int = function
  | Scalar  s ->
    begin match s with
    | Snative_int i -> i
    | Sstring _
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sint _        | Sint_couple _        | Sint_triple _
                    | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "native_int" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "native_int" t

and to_native_int_couple = function
  | Scalar  s ->
    begin match s with
    | Snative_int_couple ic -> ic
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _                        | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "native_int_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "native_int_couple" t

and to_native_int_triple = function
  | Scalar  s ->
    begin match s with
    | Snative_int_triple it -> it
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "native_int_triple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "native_int_triple" t

and to_int32 = function
  | Scalar  s ->
    begin match s with
    | Sint32 i -> i
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
                    | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int32" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int32" t

and to_int32_couple = function
  | Scalar  s ->
    begin match s with
    | Sint32_couple ic -> ic
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _                             | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int32_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int32_couple" t

and to_int32_triple = function
  | Scalar  s ->
    begin match s with
    | Sint32_triple ic -> ic
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int32_triple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int32_triple" t

and to_int64 = function
  | Scalar  s ->
    begin match s with
    | Sint64 i -> i
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
                    | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int64" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int64" t

and to_int64_couple = function
  | Scalar  s ->
    begin match s with
    | Sint64_couple ic -> ic
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _                             | Sint64_triple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int64_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int64_couple" t

and to_int64_triple = function
  | Scalar  s ->
    begin match s with
    | Sint64_triple ic -> ic
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _
    | Sfloat _      | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "int64_triple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "int64_triple" t

and to_float = function
  | Scalar  s ->
    begin match s with
    | Sfloat f -> f
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
                    | Sfloat_couple _      | Sfloat_triple _ as s ->
      raise_wrong_communication "float" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "float" t

and to_float_couple = function
  | Scalar  s ->
    begin match s with
    | Sfloat_couple fc -> fc
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _                             | Sfloat_triple _ as s ->
      raise_wrong_communication "float_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "float_couple" t

and to_float_triple = function
  | Scalar  s ->
    begin match s with
    | Sfloat_triple ft -> ft
    | Sbool _       | Sbool_couple _       | Sbool_triple _
    | Sstring _
    | Sint _        | Sint_couple _        | Sint_triple _
    | Snative_int _ | Snative_int_couple _ | Snative_int_triple _
    | Sint32 _      | Sint32_couple _      | Sint32_triple _
    | Sint64 _      | Sint64_couple _      | Sint64_triple _
    | Sfloat _      | Sfloat_couple _ as s ->
      raise_wrong_communication "float_couple" (Scalar s) end
  | Vector _ | Matrix _ as t ->
    raise_wrong_communication "float_triple" t

and to_bool_vector = function
  | Vector v ->
    begin match v with
    | Vbool bv -> bv
                    | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "bool_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "bool_vector" t

and to_bool_couple_vector = function
  | Vector v ->
    begin match v with
    | Vbool_couple bcv -> bcv
    | Vbool _                              | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "bool_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "bool_couple_vector" t

and to_bool_triple_vector = function
  | Vector v ->
    begin match v with
    | Vbool_triple btv -> btv
    | Vbool _       | Vbool_couple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "bool_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "bool_triple_vector" t

and to_string_vector = function
  | Vector v ->
    begin match v with
    | Vstring sv -> sv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "string_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "string_vector" t

and to_int_vector = function
  | Vector v ->
    begin match v with
    | Vint iv -> iv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
                    | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int_vector" t

and to_int_couple_vector = function
  | Vector v ->
    begin match v with
    | Vint_couple icv -> icv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _                               | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int_couple_vector" t

and to_int_triple_vector = function
  | Vector v ->
    begin match v with
    | Vint_triple itv -> itv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int_triple_vector" t

and to_native_int_vector = function
  | Vector v ->
    begin match v with
    | Vnative_int iv -> iv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
                    | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "native_int_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "native_int_vector" t

and to_native_int_couple_vector = function
  | Vector v ->
    begin match v with
    | Vnative_int_couple icv -> icv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _                        | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "native_int_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "native_int_couple_vector" t

and to_native_int_triple_vector = function
  | Vector v ->
    begin match v with
    | Vnative_int_triple itv -> itv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "native_int_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "native_int_triple_vector" t

and to_int32_vector = function
  | Vector v ->
    begin match v with
    | Vint32 iv -> iv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
                    | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int32_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int32_vector" t

and to_int32_couple_vector = function
  | Vector v ->
    begin match v with
    | Vint32_couple icv -> icv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _                             | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int32_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int32_couple_vector" t

and to_int32_triple_vector = function
  | Vector v ->
    begin match v with
    | Vint32_triple itv -> itv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int32_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int32_triple_vector" t

and to_int64_vector = function
  | Vector v ->
    begin match v with
    | Vint64 iv -> iv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
                    | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int64_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int64_vector" t

and to_int64_couple_vector = function
  | Vector v ->
    begin match v with
    | Vint64_couple icv -> icv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _                             | Vint64_triple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int64_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int64_couple_vector" t

and to_int64_triple_vector = function
  | Vector v ->
    begin match v with
    | Vint64_triple itv -> itv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _
    | Vfloat _      | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "int64_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "int64_triple_vector" t

and to_float_vector = function
  | Vector v ->
    begin match v with
    | Vfloat fv -> fv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
                    | Vfloat_couple _      | Vfloat_triple _ as v ->
      raise_wrong_communication "float_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "float_vector" t

and to_float_couple_vector = function
  | Vector v ->
    begin match v with
    | Vfloat_couple fcv -> fcv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _                             | Vfloat_triple _ as v ->
      raise_wrong_communication "float_couple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "float_couple_vector" t

and to_float_triple_vector = function
  | Vector v ->
    begin match v with
    | Vfloat_triple ftv -> ftv
    | Vbool _       | Vbool_couple _       | Vbool_triple _
    | Vstring _
    | Vint _        | Vint_couple _        | Vint_triple _
    | Vnative_int _ | Vnative_int_couple _ | Vnative_int_triple _
    | Vint32 _      | Vint32_couple _      | Vint32_triple _
    | Vint64 _      | Vint64_couple _      | Vint64_triple _
    | Vfloat _      | Vfloat_couple _ as v ->
      raise_wrong_communication "float_triple_vector" (Vector v) end
  | Scalar _ | Matrix _ as t ->
    raise_wrong_communication "float_triple_vector" t

and to_c_float_matrix = function
  | Matrix m ->
    begin match m with
    | Mcfloat cfm -> cfm
    | Mffloat _ as m ->
      raise_wrong_communication "c_float_matrix" (Matrix m) end
  | Scalar _ | Vector _ as t ->
    raise_wrong_communication "c_float_matrix" t

and to_fortran_float_matrix = function
  | Matrix m ->
    begin match m with
    | Mffloat ffm -> ffm
    | Mcfloat _ as m ->
      raise_wrong_communication "fortran_float_matrix" (Matrix m) end
  | Scalar _ | Vector _ as t ->
    raise_wrong_communication "fortran_float_matrix" t
;;

let of_bool b = Scalar (Sbool b)
and of_bool_couple bc = Scalar (Sbool_couple bc)
and of_bool_triple bt = Scalar (Sbool_triple bt)
and of_string s = Scalar (Sstring s)
and of_int i = Scalar (Sint i)
and of_int_couple ic = Scalar (Sint_couple ic)
and of_int_triple it = Scalar (Sint_triple it)
and of_native_int i = Scalar (Snative_int i)
and of_native_int_couple ic = Scalar (Snative_int_couple ic)
and of_native_int_triple it = Scalar (Snative_int_triple it)
and of_int32 i = Scalar (Sint32 i)
and of_int32_couple ic = Scalar (Sint32_couple ic)
and of_int32_triple it = Scalar (Sint32_triple it)
and of_int64 i = Scalar (Sint64 i)
and of_int64_couple ic = Scalar (Sint64_couple ic)
and of_int64_triple it = Scalar (Sint64_triple it)
and of_float f = Scalar (Sfloat f)
and of_float_couple fc = Scalar (Sfloat_couple fc)
and of_float_triple ft = Scalar (Sfloat_triple ft)
and of_bool_vector bv = Vector (Vbool bv)
and of_bool_couple_vector bcv = Vector (Vbool_couple bcv)
and of_bool_triple_vector btv = Vector (Vbool_triple btv)
and of_string_vector sv = Vector (Vstring sv)
and of_int_vector iv = Vector (Vint iv)
and of_int_couple_vector icv = Vector (Vint_couple icv)
and of_int_triple_vector itv = Vector (Vint_triple itv)
and of_native_int_vector iv = Vector (Vnative_int iv)
and of_native_int_couple_vector icv = Vector (Vnative_int_couple icv)
and of_native_int_triple_vector itv = Vector (Vnative_int_triple itv)
and of_int32_vector iv = Vector (Vint32 iv)
and of_int32_couple_vector icv = Vector (Vint32_couple icv)
and of_int32_triple_vector itv = Vector (Vint32_triple itv)
and of_int64_vector iv = Vector (Vint64 iv)
and of_int64_couple_vector icv = Vector (Vint64_couple icv)
and of_int64_triple_vector itv = Vector (Vint64_triple itv)
and of_float_vector fv = Vector (Vfloat fv)
and of_float_couple_vector fcv = Vector (Vfloat_couple fcv)
and of_float_triple_vector ftv = Vector (Vfloat_triple ftv)
and of_c_float_matrix cfm = Matrix (Mcfloat cfm)
and of_fortran_float_matrix ffm = Matrix (Mffloat ffm)
;;
