(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parfun.ml,v 1.43 2013/04/03 12:21:03 fclement Exp $ *)

exception Parfun_error;;

type ('x, 'a, 'b) worker = {
   io : IO.ic * IO.oc;
   f : 'a Sklstream.t -> 'b Sklstream.t;
};;

type master = { daemon_io : Skl_comm.iochans; };;

type ('x, 'a, 'b) startup_ret =
   | Worker of ('x, 'a, 'b) worker
   | Master of master
;;

let request_daemon io req =
  match
    Skl_comm.request_daemon io req
  with
  | Some (Skl_comm.Success addr) ->
    addr
  | Some Skl_comm.Failure | None ->
    raise Parfun_error
;;

(* Core of the parallel implementation *)

let rec start_tree :
'a 'b. Skl_comm.iochans -> ('a, 'b) Skl.skel -> Iotree.ioatree =
  fun io tree ->
    match tree with
    | Sk.Skl.Par par ->
      Sk.with_par par {
        Sk.f_par = fun t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          Iotree.IOABranch (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Rails rails ->
      Sk.with_rails rails {
        Sk.f_rails = fun ska ->
          let addr_treev =
            Array.map (start_tree io) ska
          in
          Iotree.IOALeafs addr_treev
      }
    | Sk.Skl.Sum sum ->
      Sk.with_sum sum {
        Sk.f_sum = fun _ t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          Iotree.IOASumSplit (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Pipe pipe ->
      Sk.with_pipe pipe {
        Sk.f_pipe = fun t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          let spawn_request =
            Skl_comm.Pipe_req
              (addr_tree1, addr_tree2) in
          let addr = request_daemon io spawn_request in
          Iotree.IOALeaf addr
      }
    | Sk.Skl.Atom atom ->
      Sk.with_atom atom {
        Sk.f_atom = fun params llexp color ->
          let spawn_request =
            let (p, id) = Expor.get_exportable (params, llexp) in
            Skl_comm.Atom_req
              (p, id, color) in
          let addr = request_daemon io spawn_request in
          Iotree.IOALeaf addr
      }
    | Sk.Skl.Id id_skl ->
      Sk.with_id id_skl {
        Sk.f_id = fun _ ->
          Iotree.IOALeafId
      }
    | Sk.Skl.Prod prod ->
      Sk.with_prod prod {
        Sk.f_prod = fun _ _ skl1 skl2 ->
          let addr_tree1 = start_tree io skl1 in
          let addr_tree2 = start_tree io skl2 in
          Iotree.IOASplit (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Loop loop ->
      Sk.with_loop loop {
        Sk.f_loop = fun _ ptree tree ->
          let addr_tree1 = start_tree io ptree in
          let addr_tree2 = start_tree io tree in
          let spawn_request =
            Skl_comm.Loop_req
              (addr_tree1, addr_tree2) in
          let addr = request_daemon io spawn_request in
          Iotree.IOALeaf addr
      }
;;

type parfun =
  { parfun : 'a 'b. ('a, 'b) Skl.skel -> 'a Sklstream.t -> 'b Sklstream.t; }
;;

let parfun io skl is =
  let tree =
    Iotree.iotree_of_ioatree (* Connects the tree *)
      (start_tree io skl) in
  match IO.fork () with
  | IO.Parent _ -> Iotree.receive_in_tree tree
  | IO.Child ->
    Iotree.send_in_tree tree is;
    exit 0
;;

type arg_master = {
   addr : Unix.sockaddr;
};;

type arg_worker = {
   f_id : int;
};;

type parsed_args =
   | Arg_master of arg_master
   | Arg_worker of arg_worker
;;

let (get_mode, set_master_mode, set_slave_mode) =
  let { Uaref.get = get_mode; set = set_mode; } =
    Uaref.make_aref_mandatory
      "Sklml parallel mode"
      (Arg_worker { f_id = -1; }) in
  let set_master_mode x =
    set_mode (Arg_master { addr = IO.read_sockaddr x; }) in
  let set_slave_mode x = set_mode (Arg_worker { f_id = x; }) in
  (get_mode, set_master_mode, set_slave_mode)
;;

let add_sklml_options () =
  Ucommand_option.library_add_all
    [
     Ucommand_option.of_regular
       { Ucommand_option.key = "-master";
         spec = Arg.String set_master_mode;
         doc = ""; };
     Ucommand_option.of_regular
       { Ucommand_option.key = "-atom";
         spec = Arg.Int set_slave_mode;
         doc = ""; };
    ];
;;

let phase_sklml_par = Ucommand_option.Phase "Sklml parallel";;

let add_sklml_par_phase () =
  Ucommand_option.phase_add_all phase_sklml_par [ "-master"; "-atom"; ];
;;

let fail m = Log.say Log.Err "%s" m; exit 2;;

let parse () =
  try
    add_sklml_options ();
    add_sklml_par_phase ();
    Ucommand_option.parse_phase phase_sklml_par;
  with
  | Arg.Bad m | Arg.Help m | Failure m -> fail m
;;

let startup () =
  parse ();
  let mode = try get_mode () with Failure m -> fail m in
  match mode with
  | Arg_master { addr = addr; } ->
    let io = IO.connect addr in
    Master { daemon_io = io; }
  | Arg_worker { f_id = f_id; } ->
    let f_gen =
      (Expor.unsafe_retreive f_id) in
    let (fifo_in, fifo_out) =
      Skl_comm.open_skl_fifo false (Unix.getpid ()) in
    let params = IO.of_channel fifo_in in
    let f = f_gen params in
    let (sock, addr) = IO.create_bound_socket None in
    let port = IO.get_port addr in
    Unix.listen sock 1; (* listen before giving the port *)
    IO.to_channel fifo_out port true;
    let io, _ = IO.wait_connection sock true in
    Worker { io = io; f = f; }
;;
(** Startup function, if -atom N is spotted in [Sys.argv] then [startup ()]
 runs the atom code, otherwise, starts as master and connects to the Sklml
 parallel launcher daemon given as argument. *)

let run_worker w =
  let { io = (ic, oc); f = f; } = w in
  Sklstream.to_io_channel (f (Sklstream.of_io_channel ic)) oc
;;

let pardo f =
  match startup () with
  | Master { daemon_io = io; } ->
    f { parfun = fun x -> parfun io x; }
  | Worker w ->
    run_worker w;
    Log.say (Log.Log 0) "Worker job done."
;;
(** Entry point. *)
