(* $Id: tester.ml,v 1.3 2010/10/15 15:45:26 qcarbonn Exp $ *)

(* Simple tester *)

(* Helpers *)

let read_sockaddr str =
  let sb = Scanf.Scanning.from_string str in
  try
    let addr, port =
      Scanf.bscanf sb "%s@:%d"
        (fun a p ->
          (Unix.inet_addr_of_string a), p) in
    Unix.ADDR_INET (addr, port)
  with _ -> failwith "Invalid address format, use www.xxx.yyy.zzz:port format"
;;

let print_sockaddr oc addr =
  match addr with
  | Unix.ADDR_UNIX addr -> Printf.fprintf oc "%s" addr
  | Unix.ADDR_INET (addr, port) ->
    Printf.fprintf oc "%s:%d" (Unix.string_of_inet_addr addr) port
;;
(** Display a sockaddr [addr] in the output channel [oc]. *)

let create_socket () =
  Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0
;;

let create_bound_socket () =
  let sock = create_socket () in
  Unix.bind sock (Unix.ADDR_INET (Unix.inet_addr_any, 0));
  (sock, Unix.getsockname sock)
;;

let create_connected_socket addr =
  let sock = create_socket () in
  Unix.connect sock addr;
  sock
;;

(* Main code *)

type action =
  | Listenner of Unix.sockaddr
  | Finite_sender of int
  | Infinite_sender
  | Unspecified
;;

let run_listenner addr =
  let sock = create_connected_socket addr in
  Printf.printf "Connected to %a.\n" print_sockaddr addr;
  Printf.printf "\t^C to escape\n\n";
  Printf.printf "Receiving data.\n"; flush stdout;
  let ic = Unix.in_channel_of_descr sock in
  let rec receive () =
    match Marshal.from_channel ic with
    | None -> ()
    | Some i -> Printf.printf "Received: Some %i\n" i; receive () in
  receive ();
  Printf.printf "Received: None\nbye.\n"
;;

let run_sender n =
  let sock, addr = create_bound_socket () in
  Printf.printf "Listenning on %a ...\n" print_sockaddr addr;
  Printf.printf "\t^C to escape\n\n"; flush stdout;
  Unix.listen sock 3;
  let (client_sock, client_addr) = Unix.accept sock in
  Printf.printf
    "Connected to %a, start sending.\n" print_sockaddr client_addr;
  let oc = Unix.out_channel_of_descr client_sock in
  match n with
  | Some n ->
    let rec loop c =
      if c > n then Marshal.to_channel oc None [] else
      begin
        Marshal.to_channel oc (Some c) [];
        Printf.printf "Sending: Some %d\n" c;
        loop (c + 1)
      end in
    loop 0;
    Printf.printf "Sending: None\nbye.\n";
  | None ->
    let c = ref 0 in
    while true do
      Marshal.to_channel oc (Some !c) [];
      Printf.printf "Sending: Some %d\n" !c;
      c := !c + 1
    done
;;

let get_usage () =
  "tester: [OPTIONS] MODE"
;;

let parse_args argv =
  let stream_size = ref 10 in
  let (listen_builder, fsend_builder, isend_builder, get_mode) =
    let mode = ref Unspecified in
    (fun addr -> mode := Listenner (read_sockaddr addr)),
    (fun () -> mode := Finite_sender !stream_size),
    (fun () -> mode := Infinite_sender),
    (fun () -> !mode) in
  let arglist = Arg.align
    [ "-n", Arg.Set_int stream_size,
        "N Sets the number of integers to send (use it before -fs)";
      "-l", Arg.String listen_builder,
        "ADDR Runs the tester in `listenner' mode";
      "-fs", Arg.Unit fsend_builder,
        " Runs the tester in `finite sender' mode";
      "-is", Arg.Unit isend_builder,
       " Runs the tester in `infinite sender' mode"; ] in
  (try
    Arg.parse_argv argv arglist ignore (get_usage ());
  with
    | Arg.Bad msg -> Printf.printf "%s" msg; exit 2
    | Arg.Help msg -> Printf.printf "%s" msg; exit 2);
  let mode = get_mode () in
  mode
;;

let main () =
  let mode = parse_args Sys.argv in
  match mode with
  | Unspecified -> failwith "Mode unspecified"
  | Listenner addr -> run_listenner addr
  | Finite_sender n -> run_sender (Some n)
  | Infinite_sender -> run_sender None
;;

Printexc.catch main ();;
